

class TetrisAI implements Runnable {
    Tetris tetris;
    boolean next;
    int[][] board;
    int px, py, pr, piece;

    public TetrisAI(Tetris t) {
        tetris=t;
        board=tetris.getBoard();
    }
    public void start() {
        new Thread(this).start();
    }



    public synchronized void reset() {
        //notifyAll();
    }

    public synchronized void next() { 
        next=true;
        notifyAll();
    }

    public synchronized void waitForNext() {
        try { 
            while(!next) wait();
        } catch(InterruptedException e) {
            e.printStackTrace(System.err);
        }
        next=false;
    }
    private boolean get(int x, int y) {
        if(x<px || x>=px+4 || y<py || y>=py+4) 
            return board[y][x]>0;

        if(Tetris.pieces[piece][pr][(y-py)*4+(x-px)]!=0) return true;
        return board[y][x]>0;
    }

    public void run() {
        while(true) {
            int maxscore=Integer.MIN_VALUE;
            int maxpx=0, maxpr=0;

            waitForNext();
            px=pr=-13;
            piece=tetris.getPiece();
            
            for(px=-4; px<board[0].length+4 && !next; px++) {
                for(pr=0; pr<Tetris.pieces[piece].length && !next; pr++) {
                    if(!tetris.spaceFor(px, 0, Tetris.pieces[piece][pr])) {
                        continue;
                    }

                    py=0;
                    while(!tetris.parkPiece(px,py,Tetris.pieces[piece][pr])) py++;

                    int n=evaluate();
                    if(n>maxscore) { 
                        maxscore=n;
                        maxpx=px;
                        maxpr=pr;
                    }
                }
            }

            for(int i=0; i<4 && tetris.getRotation()!=maxpr; i++)
                tetris.rotate(true);

            System.out.println("Max score: "+maxscore+", "+maxpx+"@"+maxpr+": "+next);

            int displace=maxpx-tetris.getPieceX();
            System.out.println("Displace: "+displace);
            while(displace>0) {
                if(!tetris.right()) System.out.println("Can't move to "+maxpx+" stuck on "+tetris.getPieceX());
                displace--;
            }
            while(displace<0) {
                if(!tetris.left()) System.out.println("Can't move to "+maxpx+" stuck on "+tetris.getPieceX());
                displace++;
            }
            if(tetris.getPieceX()!=maxpx) {
                System.out.println("Uh oh");
            } else {
                tetris.drop();
            }
        }
    }
    int evaluate() { 
        int r=evaluate2();
        System.out.println("Evaluating: "+r);
        return r;
    }
    int evaluate2() {
        int sum=0; 
        /*
        switch(countRows()) {
            case 1: sum+=1; break;
            case 2: sum+=500; break;
            case 3: sum+=5000; break;
            case 4: sum+=100000; break;
        }
        */
        sum+=ironMill();
        sum-=countCoveredHoles()*10000;
        return sum;
    }
    
    int evaluate1() {
        int sum=0; 
        switch(countRows()) {
            case 1: sum+=1; break;
            case 2: sum+=500; break;
            case 3: sum+=5000; break;
            case 4: sum+=100000; break;
        }
        //sum-=countCoveredHoles();
        //sum+=density();
        sum+=-countHeight()*2-countCoveredHoles()*5;
        return sum;
    }

    int countRows() { 
        int r=0;

        for(int y=0; y<board.length; y++) {
            boolean ok=true;
            for(int x=0; x<board[y].length && ok; x++) 
                if(!get(x,y)) ok=false;
            if(ok) r++;
        }
        return r;
    }

    int countCoveredHoles() {
        int holes=0;
        for(int x=0; x<board[0].length; x++) {
            boolean swap=false;
            for(int y=0; y<board.length; y++) {
                if(get(x,y)) swap=true;
                else { 
                    if(swap) holes++;
                    swap=false;
                }
            }
        }
        System.out.println("Holes: "+holes);
        return holes;
    }

    int countHeight() {
        for(int y=0; y<board.length; y++) {
            for(int x=0; x<board[0].length; x++) {
                if(get(x,y)) return board.length-y;
            }
        }
        return 0;
    }

    int ironMill() { 
        int iron=0;
        for(int y=0; y<board.length; y++) {
            for(int x=0; x<board[0].length; x++) {
                if(get(x,y)) iron+=y*y*y;
            }
        }
        return iron;
    }

    int density() { 
        int density=0;
        for(int y=0; y<board.length; y++) {
            int num=0;
            for(int x=0; x<board[0].length; x++) {
                if(get(x,y)) num++;
            }
            int p=(num*100/board.length);
            density+=p;
        }
        return density/countHeight();
    }

}

